from typing import List
from typing import Optional

from .base_models import CovJsonBaseModel
from .i18n import i18n


class Category(CovJsonBaseModel):
    id: str
    label: i18n
    description: Optional[i18n] = None


class ObservedProperty(CovJsonBaseModel):
    id: Optional[str] = None
    label: i18n
    description: Optional[i18n] = None
    categories: Optional[List[Category]] = None



import json
import sys
from io import StringIO
from pathlib import Path

import pytest
from covjson_pydantic.coverage import Coverage
from covjson_pydantic.coverage import CoverageCollection
from covjson_pydantic.domain import Axes
from covjson_pydantic.domain import Domain
from covjson_pydantic.ndarray import NdArray
from covjson_pydantic.ndarray import NdArrayFloat
from covjson_pydantic.ndarray import NdArrayInt
from covjson_pydantic.ndarray import NdArrayStr
from covjson_pydantic.ndarray import TiledNdArrayFloat
from covjson_pydantic.parameter import Parameter
from covjson_pydantic.parameter import ParameterGroup
from covjson_pydantic.reference_system import ReferenceSystem
from pydantic import ValidationError


happy_cases = [
    ("spec-axes.json", Axes),
    ("str-axes.json", Axes),
    ("coverage-json.json", Coverage),
    ("coverage-mixed-type-ndarray.json", Coverage),
    ("doc-example-coverage.json", Coverage),
    ("example_py.json", Coverage),
    ("spec-vertical-profile-coverage.json", Coverage),
    ("spec-trajectory-coverage.json", Coverage),
    ("doc-example-coverage-collection.json", CoverageCollection),
    ("grid-domain.json", Domain),
    ("point-series-domain-custom.json", Domain),
    ("spec-domain-grid.json", Domain),
    ("spec-domain-vertical-profile.json", Domain),
    ("spec-domain-point-series.json", Domain),
    ("spec-domain-point.json", Domain),
    ("spec-domain-point-compact.json", Domain),
    ("spec-domain-multipoint-series.json", Domain),
    ("spec-domain-multipoint.json", Domain),
    ("spec-domain-trajectory.json", Domain),
    ("ndarray-float.json", NdArrayFloat),
    ("ndarray-string.json", NdArrayStr),
    ("ndarray-integer.json", NdArrayInt),
    ("spec-ndarray.json", NdArrayFloat),
    ("spec-tiled-ndarray.json", TiledNdArrayFloat),
    ("continuous-data-parameter.json", Parameter),
    ("categorical-data-parameter.json", Parameter),
    ("spec-parametergroup.json", ParameterGroup),
    ("spec-reference-system-identifierrs.json", ReferenceSystem),
]


@pytest.mark.parametrize("file_name, object_type", happy_cases)
def test_happy_cases(file_name, object_type):
    file = Path(__file__).parent.resolve() / "test_data" / file_name
    # Put JSON in default unindented format
    with open(file, "r") as f:
        data = json.load(f)
    json_string = json.dumps(data, separators=(",", ":"))

    # Round-trip
    assert object_type.model_validate_json(json_string).model_dump_json(exclude_none=True) == json_string


error_cases = [
    ("grid-domain-no-y.json", Domain, r"A 'Grid' must have a 'y'-axis"),
    (
        "point-series-domain-more-z.json",
        Domain,
        r"If provided, the 'values' field of the ValuesAxis 'z'-axis of a 'PointSeries' "
        + "domain must contain a single value.",
    ),
    ("point-series-domain-no-t.json", Domain, r"A 'PointSeries' must have a 't'-axis."),
    ("mixed-type-axes.json", Axes, r"Input should be a valid number"),
    ("mixed-type-axes-2.json", Axes, r"Input should be a valid string"),
    ("mixed-type-ndarray-1.json", NdArrayFloat, r"Input should be a valid number"),
    ("mixed-type-ndarray-1.json", NdArrayStr, r"Input should be 'string'"),
    ("mixed-type-ndarray-2.json", NdArrayFloat, r"Input should be a valid number"),
    ("mixed-type-ndarray-2.json", NdArrayStr, r"Input should be 'string'"),
    ("mixed-type-ndarray-3.json", NdArrayInt, r"Input should be a valid integer"),
    ("mixed-type-ndarray-3.json", NdArrayFloat, r"Input should be 'float'"),
]


@pytest.mark.parametrize("file_name, object_type, error_message", error_cases)
def test_error_cases(file_name, object_type, error_message):
    file = Path(__file__).parent.resolve() / "test_data" / file_name
    # Put JSON in default unindented format
    with open(file, "r") as f:
        data = json.load(f)
    json_string = json.dumps(data, separators=(",", ":"))

    with pytest.raises(ValidationError, match=error_message):
        object_type.model_validate_json(json_string)


def test_ndarray_directly():
    with pytest.raises(TypeError, match="NdArray cannot be instantiated directly"):
        NdArray(axisNames=["x", "y", "t"], shape=[1, 1, 1], values=[42.0])


def test_example_py():
    file = Path(__file__).parent.parent.resolve() / "example.py"

    with open(file, "r") as f:
        code = f.read()

    old_stdout = sys.stdout
    sys.stdout = my_stdout = StringIO()
    exec(code)
    sys.stdout = old_stdout

    file = Path(__file__).parent.resolve() / "test_data" / "example_py.json"
    with open(file, "r") as f:
        assert my_stdout.getvalue() == f.read()



import math
from typing import List
from typing import Literal
from typing import Optional

from pydantic import model_validator

from .base_models import CovJsonBaseModel


class NdArray(CovJsonBaseModel, extra="allow"):
    type: Literal["NdArray"] = "NdArray"
    dataType: str  # Kept here to ensure order of output in JSON  # noqa: N815
    axisNames: Optional[List[str]] = None  # noqa: N815
    shape: Optional[List[int]] = None

    @model_validator(mode="before")
    @classmethod
    def validate_is_sub_class(cls, values):
        if cls is NdArray:
            raise TypeError(
                "NdArray cannot be instantiated directly, please use a NdArrayFloat, NdArrayInt or NdArrayStr"
            )
        return values

    @model_validator(mode="after")
    def check_field_dependencies(self):
        if len(self.values) > 1 and (self.axisNames is None or len(self.axisNames) == 0):
            raise ValueError("'axisNames' must to be provided if array is not 0D")

        if len(self.values) > 1 and (self.shape is None or len(self.shape) == 0):
            raise ValueError("'shape' must to be provided if array is not 0D")

        if self.axisNames is not None and self.shape is not None and len(self.axisNames) != len(self.shape):
            raise ValueError("'axisNames' and 'shape' should have equal length")

        if self.shape is not None and len(self.shape) >= 1:
            prod = math.prod(self.shape)
            if len(self.values) != prod:
                raise ValueError(
                    "Where 'shape' is present and non-empty, the product of its values MUST equal "
                    "the number of elements in the 'values' array."
                )

        return self


class NdArrayFloat(NdArray):
    dataType: Literal["float"] = "float"  # noqa: N815
    values: List[Optional[float]]


class NdArrayInt(NdArray):
    dataType: Literal["integer"] = "integer"  # noqa: N815
    values: List[Optional[int]]


class NdArrayStr(NdArray):
    dataType: Literal["string"] = "string"  # noqa: N815
    values: List[Optional[str]]


class TileSet(CovJsonBaseModel):
    tileShape: List[Optional[int]]  # noqa: N815
    urlTemplate: str  # noqa: N815


# TODO: Validation of field dependencies
# TODO: Support string and integer type TiledNdArray
class TiledNdArrayFloat(CovJsonBaseModel, extra="allow"):
    type: Literal["TiledNdArray"] = "TiledNdArray"
    dataType: Literal["float"] = "float"  # noqa: N815
    axisNames: List[str]  # noqa: N815
    shape: List[int]
    tileSets: List[TileSet]  # noqa: N815



from enum import Enum
from typing import Dict


class LanguageTag(str, Enum):
    dutch = "nl"
    english = "en"
    german = "de"
    undefined = "und"


# TODO: This was throwing warning:
#   Expected `definition-ref` but got `LanguageTag` - serialized value may not be as expected
# This may be a bug in Pydantic: https://github.com/pydantic/pydantic/issues/6467
# or: https://github.com/pydantic/pydantic/issues/6422
# So, for now, reverted to a less strict type
# See issue: https://github.com/KNMI/covjson-pydantic/issues/3
# i18n = Dict[LanguageTag, str]
i18n = Dict[str, str]



import sys

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from pydantic import AnyUrl
from pydantic import Field

from .base_models import CovJsonBaseModel
from .domain import Domain
from .domain import DomainType
from .ndarray import NdArrayFloat
from .ndarray import NdArrayInt
from .ndarray import NdArrayStr
from .ndarray import TiledNdArrayFloat
from .parameter import Parameter
from .parameter import ParameterGroup
from .reference_system import ReferenceSystemConnectionObject

NdArrayTypes = Annotated[Union[NdArrayFloat, NdArrayInt, NdArrayStr], Field(discriminator="dataType")]


class Coverage(CovJsonBaseModel, extra="allow"):
    id: Optional[str] = None
    type: Literal["Coverage"] = "Coverage"
    domain: Domain
    parameters: Optional[Dict[str, Parameter]] = None
    parameterGroups: Optional[List[ParameterGroup]] = None  # noqa: N815
    ranges: Dict[str, Union[NdArrayTypes, TiledNdArrayFloat, AnyUrl]]


class CoverageCollection(CovJsonBaseModel, extra="allow"):
    type: Literal["CoverageCollection"] = "CoverageCollection"
    domainType: Optional[DomainType] = None  # noqa: N815
    coverages: List[Coverage]
    parameters: Optional[Dict[str, Parameter]] = None
    parameterGroups: Optional[List[ParameterGroup]] = None  # noqa: N815
    referencing: Optional[List[ReferenceSystemConnectionObject]] = None



from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict


class CovJsonBaseModel(PydanticBaseModel):
    model_config = ConfigDict(
        str_strip_whitespace=True,
        str_min_length=1,
        extra="forbid",
        validate_default=True,
        validate_assignment=True,
        strict=True,
    )



import timeit
from pathlib import Path

filename = Path(__file__).parent.resolve() / "tests" / "test_data" / "coverage-json.json"

setup = f"""
import json
from covjson_pydantic.coverage import Coverage

file = "{filename}"
# Put JSON in default unindented format
with open(file, "r") as f:
    data = json.load(f)
json_string = json.dumps(data, separators=(",", ":"))
cj = Coverage.model_validate_json(json_string)
"""

# This can be used to quickly check performance. The first call checks JSON to Python conversion
# The second call checks Python to JSON conversion
# Consider generating a larger CoverageJSON file
print(timeit.timeit("Coverage.model_validate_json(json_string)", setup, number=1000))
print(timeit.timeit("cj.model_dump_json(exclude_none=True)", setup, number=1000))



from enum import Enum
from typing import Generic
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

from pydantic import AwareDatetime
from pydantic import field_validator
from pydantic import model_validator
from pydantic import PositiveInt

from .base_models import CovJsonBaseModel
from .reference_system import ReferenceSystemConnectionObject


class CompactAxis(CovJsonBaseModel):
    start: float
    stop: float
    num: PositiveInt

    @model_validator(mode="after")
    def single_value_case(self):
        if self.num == 1 and self.start != self.stop:
            raise ValueError("If the value of 'num' is 1, then 'start' and 'stop' MUST have identical values.")
        return self


ValuesT = TypeVar("ValuesT")


# Combination between Generics (ValuesT) and datetime and strict mode causes issues between JSON <-> Pydantic
# conversions. Strict mode has been disabled. Issue: https://github.com/KNMI/covjson-pydantic/issues/4
class ValuesAxis(CovJsonBaseModel, Generic[ValuesT], extra="allow", strict=False):
    dataType: Optional[str] = None  # noqa: N815
    coordinates: Optional[List[str]] = None
    values: List[ValuesT]
    bounds: Optional[List[ValuesT]] = None

    @model_validator(mode="after")
    def bounds_length(self):
        if self.bounds is not None and len(self.bounds) != 2 * len(self.values):
            raise ValueError("If provided, the length of 'bounds' should be twice that of 'values'.")
        return self


class DomainType(str, Enum):
    grid = "Grid"
    vertical_profile = "VerticalProfile"
    point_series = "PointSeries"
    point = "Point"
    multi_point_series = "MultiPointSeries"
    multi_point = "MultiPoint"
    trajectory = "Trajectory"


class Axes(CovJsonBaseModel):
    x: Optional[Union[ValuesAxis[float], ValuesAxis[str], CompactAxis]] = None
    y: Optional[Union[ValuesAxis[float], ValuesAxis[str], CompactAxis]] = None
    z: Optional[Union[ValuesAxis[float], ValuesAxis[str], CompactAxis]] = None
    t: Optional[ValuesAxis[AwareDatetime]] = None
    composite: Optional[ValuesAxis[Tuple]] = None

    @model_validator(mode="after")
    def at_least_one_axes(self):
        if self.x is None and self.y is None and self.z is None and self.t is None and self.composite is None:
            raise ValueError("At least one axis of x,y,z,t or composite must be given.")
        return self


class Domain(CovJsonBaseModel, extra="allow"):
    type: Literal["Domain"] = "Domain"
    domainType: Optional[DomainType] = None  # noqa: N815
    axes: Axes
    referencing: Optional[List[ReferenceSystemConnectionObject]] = None

    # TODO: This is a workaround to allow domainType to work in strict mode, in combination with FastAPI.
    # See: https://github.com/tiangolo/fastapi/discussions/9868
    # And: https://github.com/KNMI/covjson-pydantic/issues/5
    @field_validator("domainType", mode="before")
    @classmethod
    def value_to_enum(cls, v):
        if isinstance(v, str):
            return DomainType(v)
        return v

    @staticmethod
    def check_axis(domain_type, axes, required_axes, allowed_axes, single_value_axes):
        # Check required axes
        for axis_name in required_axes:
            axis = getattr(axes, axis_name)
            if axis is None:
                raise ValueError(f"A '{domain_type.value}' must have a '{axis_name}'-axis.")
            if axis_name in single_value_axes:
                if isinstance(axis, ValuesAxis) and len(axis.values) != 1:
                    raise ValueError(
                        f"The 'values' field of the ValuesAxis '{axis_name}'-axis "
                        f"of a '{domain_type.value}' domain must contain a single value."
                    )
                if isinstance(axis, CompactAxis) and axis.num != 1:
                    raise ValueError(
                        f"The 'num' field of the CompactAxis '{axis_name}'-axis "
                        f"of a '{domain_type.value}' domain must be 1."
                    )

        # Check allowed axes
        all_axis = {"x", "y", "z", "t", "composite"}
        for axis_name in all_axis - required_axes - allowed_axes:
            axis = getattr(axes, axis_name)
            if axis is not None:
                raise ValueError(f"A '{domain_type.value}' domain can not have a '{axis_name}'-axis.")

        # Check for single value of allowed axes
        for axis_name in allowed_axes:
            axis = getattr(axes, axis_name)
            if axis is not None and axis_name in single_value_axes:
                if isinstance(axis, ValuesAxis) and len(axis.values) != 1:
                    raise ValueError(
                        f"If provided, the 'values' field of the ValuesAxis '{axis_name}'-axis "
                        f"of a '{domain_type.value}' domain must contain a single value."
                    )
                if isinstance(axis, CompactAxis) and axis.num != 1:
                    raise ValueError(
                        f"If provided, the 'num' field of the CompactAxis '{axis_name}'-axis "
                        f"of a '{domain_type.value}' domain must be 1."
                    )

    @model_validator(mode="after")
    def check_domain_consistent(self):
        domain_type = self.domainType
        axes = self.axes

        if domain_type == DomainType.grid:
            Domain.check_axis(
                domain_type, axes, required_axes={"x", "y"}, allowed_axes={"z", "t"}, single_value_axes=set()
            )

        if domain_type == DomainType.vertical_profile:
            Domain.check_axis(
                domain_type, axes, required_axes={"x", "y", "z"}, allowed_axes={"t"}, single_value_axes={"x", "y", "t"}
            )

        if domain_type == DomainType.point_series:
            Domain.check_axis(
                domain_type, axes, required_axes={"x", "y", "t"}, allowed_axes={"z"}, single_value_axes={"x", "y", "z"}
            )

        if domain_type == DomainType.point:
            Domain.check_axis(
                domain_type,
                axes,
                required_axes={"x", "y"},
                allowed_axes={"z", "t"},
                single_value_axes={"x", "y", "z", "t"},
            )

        if domain_type == DomainType.multi_point_series:
            Domain.check_axis(
                domain_type, axes, required_axes={"composite", "t"}, allowed_axes=set(), single_value_axes=set()
            )

        if domain_type == DomainType.multi_point:
            Domain.check_axis(
                domain_type, axes, required_axes={"composite"}, allowed_axes={"t"}, single_value_axes={"t"}
            )

        if domain_type == DomainType.trajectory:
            Domain.check_axis(
                domain_type, axes, required_axes={"composite"}, allowed_axes={"z"}, single_value_axes={"z"}
            )

        return self



from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from pydantic import model_validator

from .base_models import CovJsonBaseModel
from .i18n import i18n
from .observed_property import ObservedProperty
from .unit import Unit


class Parameter(CovJsonBaseModel, extra="allow"):
    type: Literal["Parameter"] = "Parameter"
    id: Optional[str] = None
    label: Optional[i18n] = None
    description: Optional[i18n] = None
    observedProperty: ObservedProperty  # noqa: N815
    categoryEncoding: Optional[Dict[str, Union[int, List[int]]]] = None  # noqa: N815
    unit: Optional[Unit] = None

    @model_validator(mode="after")
    def must_not_have_unit_if_observed_property_has_categories(self):
        if self.unit is not None and self.observedProperty is not None and self.observedProperty.categories is not None:
            raise ValueError(
                "A parameter object MUST NOT have a 'unit' member "
                "if the 'observedProperty' member has a 'categories' member."
            )

        return self


class ParameterGroup(CovJsonBaseModel, extra="allow"):
    type: Literal["ParameterGroup"] = "ParameterGroup"
    id: Optional[str] = None
    label: Optional[i18n] = None
    description: Optional[i18n] = None
    observedProperty: Optional[ObservedProperty] = None  # noqa: N815
    members: List[str]

    @model_validator(mode="after")
    def must_have_label_and_or_observed_property(self):
        if self.label is None and self.observedProperty is None:
            raise ValueError(
                "A parameter group object MUST have either or both the members 'label' or/and 'observedProperty'"
            )
        return self



from typing import Optional
from typing import Union

from pydantic import model_validator

from .base_models import CovJsonBaseModel
from .i18n import i18n


class Symbol(CovJsonBaseModel):
    value: str
    type: str


class Unit(CovJsonBaseModel):
    id: Optional[str] = None
    label: Optional[i18n] = None
    symbol: Optional[Union[str, Symbol]] = None

    @model_validator(mode="after")
    def check_either_label_or_symbol(self):
        if self.label is None and self.symbol is None:
            raise ValueError("Either 'label' or 'symbol' should be set")

        return self



from datetime import datetime
from datetime import timezone

from covjson_pydantic.coverage import Coverage
from covjson_pydantic.domain import Axes
from covjson_pydantic.domain import Domain
from covjson_pydantic.domain import DomainType
from covjson_pydantic.domain import ValuesAxis
from covjson_pydantic.ndarray import NdArrayFloat
from pydantic import AwareDatetime

c = Coverage(
    domain=Domain(
        domainType=DomainType.point_series,
        axes=Axes(
            x=ValuesAxis[float](values=[1.23]),
            y=ValuesAxis[float](values=[4.56]),
            t=ValuesAxis[AwareDatetime](values=[datetime(2024, 8, 1, tzinfo=timezone.utc)]),
        ),
    ),
    ranges={"temperature": NdArrayFloat(axisNames=["x", "y", "t"], shape=[1, 1, 1], values=[42.0])},
)

print(c.model_dump_json(exclude_none=True, indent=4))



import logging
import os

import setuptools

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(os.environ.get("LOG_LEVEL", "INFO"))

if __name__ == "__main__":
    setuptools.setup()



from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from pydantic import AnyUrl
from pydantic import model_validator

from .base_models import CovJsonBaseModel
from .i18n import i18n


class TargetConcept(CovJsonBaseModel):
    id: Optional[str] = None  # Not in spec, but needed for example in spec for 'Identifier-based Reference Systems'
    label: i18n
    description: Optional[i18n] = None


class ReferenceSystem(CovJsonBaseModel, extra="allow"):
    type: Literal["GeographicCRS", "ProjectedCRS", "VerticalCRS", "TemporalRS", "IdentifierRS"]
    id: Optional[str] = None
    description: Optional[i18n] = None

    # Only for TemporalRS
    calendar: Optional[Union[Literal["Gregorian"], AnyUrl]] = None
    timeScale: Optional[AnyUrl] = None  # noqa: N815

    # Only for IdentifierRS
    label: Optional[i18n] = None
    targetConcept: Optional[TargetConcept] = None  # noqa: N815
    identifiers: Optional[Dict[str, TargetConcept]] = None

    @model_validator(mode="after")
    def check_type_specific_fields(self):
        if self.type != "TemporalRS" and (self.calendar is not None or self.timeScale is not None):
            raise ValueError("'calendar' and 'timeScale' fields can only be used for type 'TemporalRS'")

        if self.type != "IdentifierRS" and (
            self.label is not None or self.targetConcept is not None or self.identifiers is not None
        ):
            raise ValueError(
                "'label', 'targetConcept' and 'identifiers' fields can only be used for type 'IdentifierRS'"
            )

        if self.type == "IdentifierRS" and self.targetConcept is None:
            raise ValueError("An identifier RS object MUST have a member 'targetConcept'")

        return self


class ReferenceSystemConnectionObject(CovJsonBaseModel):
    coordinates: List[str]
    system: ReferenceSystem
